/* Base on code from:
   http://archives.seul.org/libevent/users/Sep-2010/msg00050.html
*/
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <event2/event.h>
#include <event2/buffer.h>
#include <event2/http.h>
#include <event2/http_struct.h>

struct download_context
{
  struct evhttp_uri *uri;
  struct event_base *base;
  struct evhttp_connection *cn;
  struct evhttp_request *req;
};

void download_callback(struct evhttp_request *req, void *arg);
int download_renew_request(struct download_context *ctx);

void download_callback(struct evhttp_request *req, void *arg)
{
  struct download_context *ctx = (struct download_context *)arg;
  switch(req->response_code)
  {
    case HTTP_OK:
    {
        struct evbuffer* buf = evhttp_request_get_input_buffer(req);
        int len = evbuffer_get_length(buf);
        char *tmp = malloc(len+1);
        memcpy(tmp, evbuffer_pullup(buf, -1), len);
        tmp[len] = '\0';
        printf("%s\n", tmp);
        free(tmp);
        event_base_loopexit(ctx->base, 0);
        break;
    }

    case HTTP_MOVEPERM:
    case HTTP_MOVETEMP:
    {
      const char *new_location = evhttp_find_header(req->input_headers, "Location");
      struct evhttp_uri *new_uri = evhttp_uri_parse(new_location);
      evhttp_uri_free(ctx->uri);
      ctx->uri = new_uri;
      download_renew_request(ctx);
      return;
    }

    default:
      event_base_loopexit(ctx->base, 0);
      return;
  }
}

struct download_context *context_new(const char *url, struct event_base* base)
{
  struct download_context *ctx = calloc(1, sizeof(*ctx));
  ctx->uri = evhttp_uri_parse(url);
  ctx->base = base;
  download_renew_request(ctx);
  return ctx;
}

void context_free(struct download_context *ctx)
{
  evhttp_connection_free(ctx->cn);
  evhttp_uri_free(ctx->uri);
  free(ctx);
}

int download_renew_request(struct download_context *ctx)
{
  if (ctx->cn)
    evhttp_connection_free(ctx->cn);

  ctx->cn = evhttp_connection_base_new(ctx->base, NULL, 
                                       evhttp_uri_get_host(ctx->uri),
                                       evhttp_uri_get_port(ctx->uri) != -1 ? evhttp_uri_get_port(ctx->uri) : 80);

   ctx->req = evhttp_request_new(download_callback, ctx);
   evhttp_make_request(ctx->cn, ctx->req, EVHTTP_REQ_GET,
                       evhttp_uri_get_query(ctx->uri) ? evhttp_uri_get_query(ctx->uri) : "/");

   evhttp_add_header(ctx->req->output_headers,
                     "Host", evhttp_uri_get_host(ctx->uri));
   return 0;
}

void download_url(const char *url)
{
  struct event_base* base = event_base_new();
  struct download_context *ctx = context_new(url, base);
  event_base_dispatch(base);
  context_free(ctx);
  event_base_free(base);
}

int main(int argc, char *argv[])
{
  if (argc < 2)
  {
    printf("usage: %s http://example.com/\n", argv[0]);
    return 1;
  }

  download_url(argv[1]);

  return 0;
}
